# -*- coding: utf-8 -*-

import pytest
import brainpy as bp


class LIF(bp.NeuGroup):
  pass


def test_check_name():
  LIF(1, name='X_LIF')

  with pytest.raises(bp.errors.UniqueNameError):
    LIF(1, name='X_LIF')





