# -*- coding: utf-8 -*-


__all__ = [
  'easy_scan',
  'easy_forloop',
  'easy_while',
  'easy_cond',
]


def easy_scan(f, dyn_vars, out_vars):
  raise ValueError


def easy_forloop(lower, upper, body_fun, init_val):
  raise ValueError


def easy_while(cond_fun, body_fun, init_val):
  raise ValueError


def easy_cond(pred, true_fun, false_fun, operand):
  raise ValueError
